# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {}

modules = []

doc_url = "https://alexsax.github.ioomnidata-tools"

git_url = "https://github.com/alexsax/omnidata_tools/tree/main/"

def custom_doc_links(name): return None



import os
import argparse
import numpy as np
from PIL import Image
import random
import yaml
import pickle
from collections import defaultdict
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import DataLoader, ConcatDataset
from torch.utils.data.sampler import WeightedRandomSampler
from torchvision import transforms
import pytorch_lightning as pl
from pytorch_lightning import Trainer
from pytorch_lightning.callbacks import ModelCheckpoint
from einops import rearrange
import einops as ein
import torch 
from PIL import Image
import matplotlib.pyplot as plt
from matplotlib import cm
from mpl_toolkits.axes_grid1 import make_axes_locatable

from data.omnidata_dataset import OmnidataDataset, REPLICA_BUILDINGS
from data.augmentation import Augmentation
from modules.unet import UNet
from modules.midas.midas_net import MidasNet
from modules.midas.dpt_depth import DPTDepthModel
from losses import masked_l1_loss, virtual_normal_loss, midas_loss

def building_in_gso(building):
    return building.__contains__('-') and building.split('-')[0] in REPLICA_BUILDINGS

def building_in_replica(building):
    return building in REPLICA_BUILDINGS

def building_in_hypersim(building):
    return building.startswith('ai_')

def building_in_taskonomy(building):
    return building not in REPLICA_BUILDINGS and not building.startswith('ai_') and not building.__contains__('-')

def building_in_blendedMVS(building):
    return building.startswith('5')

class Depth(pl.LightningModule):
    def __init__(self, config_file, experiment_name):
        super().__init__()

        config = self.load_config(config_file)
        self.experiment_name = experiment_name
        self.pretrained_weights_path = config['pretrained_weights_path']
        self.image_size = config['image_size']
        self.batch_size = config['batch_size']
        self.gpus = config['gpus']
        self.num_workers = config['num_workers']
        self.learning_rate = config['lr']
        self.weight_decay = config['weight_decay']

        self.taskonomy_variant = config['taskonomy_variant']
        self.data_paths = config['data_paths']
        self.train_datasets = [dataset for dataset in config['train_datasets'].keys() if config['train_datasets'][dataset]]
        self.val_datasets = [dataset for dataset in config['val_datasets'].keys() if config['val_datasets'][dataset]]
        self.normalize_rgb = config['normalize_rgb']
        self.normalization_mean = config['normalization_mean']
        self.normalization_std = config['normalization_std']

        self.setup_datasets()
        self.save_debug_info_on_error = False
        self.num_val_images = config['num_val_images']
        self.val_samples = self.select_val_samples_for_datasets()
        self.last_log_step = 0
        self.log_step = config['log_step']
        self.save_dir = config['save_dir']
        self.save_top_k = config['save_top_k']

        self.aug = Augmentation()

        self.vnl_loss = virtual_normal_loss.VNL_Loss(1.0, 1.0, (self.image_size, self.image_size))
        self.midas_loss = midas_loss.MidasLoss(alpha=0.1)

        # self.model = DPTDepthModel(backbone='vitl16_384') # DPT Large
        self.model = DPTDepthModel()    # DPT Hybrid

        if self.pretrained_weights_path is not None:
            checkpoint = torch.load(self.pretrained_weights_path) #, map_location="cuda:1")
            # In case we load a checkpoint from this LightningModule
            if 'state_dict' in checkpoint:
                state_dict = {}
                for k, v in checkpoint['state_dict'].items():
                    # state_dict[k.replace('model.', '')] = v
                    state_dict[k[6:]] = v

            else:
                state_dict = checkpoint
            self.model.load_state_dict(state_dict)

    @staticmethod
    def load_config(config_file):
        with open(config_file, 'r') as stream:
            config = yaml.safe_load(stream)
        return config

    def setup_datasets(self):

        self.tasks = ['rgb', 'depth_zbuffer', 'depth_euclidean', 'mask_valid']      

        self.train_options = {}
        self.trainsets = {}
        for dataset in self.train_datasets:
            self.train_options[dataset] = OmnidataDataset.Options(
                tasks=self.tasks,
                datasets=[dataset],
                split='train',
                taskonomy_variant=self.taskonomy_variant,
                transform='DEFAULT',
                image_size=self.image_size,
                normalize_rgb=self.normalize_rgb,
                normalization_mean=self.normalization_mean,
                normalization_std=self.normalization_std,
            )   
            self.trainsets[dataset] = OmnidataDataset(options=self.train_options[dataset])
            print(f'Train set ({dataset}) contains {len(self.trainsets[dataset])} samples.')

        print('Loaded training sets...')


        self.val_options = {}
        self.valsets = {}
        for dataset in self.val_datasets:
            self.val_options[dataset] = OmnidataDataset.Options(
                split='val',
                taskonomy_variant=self.taskonomy_variant,
                tasks=self.tasks,
                datasets=[dataset],
                transform='DEFAULT',
                image_size=self.image_size,
                normalize_rgb=self.normalize_rgb,
                normalization_mean=self.normalization_mean,
                normalization_std=self.normalization_std,
            )
            self.valsets[dataset] = OmnidataDataset(options=self.val_options[dataset])
            self.valsets[dataset].randomize_order()
            print(f'Val set ({dataset}) contains {len(self.valsets[dataset])} samples.')

        print('Loaded validation sets...')


    def train_dataloader(self):
        trainsets = self.trainsets.values()
        trainsets_counts = [len(trainset) for trainset in trainsets]

        dataset_sample_count = torch.tensor(trainsets_counts)
        weights = 1. / dataset_sample_count.float()
        samples_weight = []
        for w, count in zip(weights, dataset_sample_count):
            samples_weight += [w] * count
        samples_weight = torch.tensor(samples_weight)
        
        sampler = WeightedRandomSampler(samples_weight, len(samples_weight))
        train_dataset = ConcatDataset(trainsets)
        return DataLoader(
            train_dataset, batch_size=self.batch_size, sampler=sampler, 
            num_workers=self.num_workers, pin_memory=False
        )

    def val_dataloader(self):
        val_dls = []
        for valset in self.valsets.values():
            dl = DataLoader(
                    valset, batch_size=self.batch_size, shuffle=False, 
                    num_workers=self.num_workers, pin_memory=False
                )

            val_dls.append(dl)    
        return val_dls   


    def forward(self, x):
        return self.model(x).unsqueeze(1) 

    def training_step(self, batch, batch_idx):
        res = self.shared_step(batch, train=True)
        # Logging
        self.log('train_ssi_loss', res['ssi_loss'], prog_bar=True, logger=True, sync_dist=len(self.gpus)>1)
        self.log('train_reg_loss', res['reg_loss'], prog_bar=True, logger=True, sync_dist=len(self.gpus)>1)
        self.log('train_vn_loss', res['vn_loss'], prog_bar=True, logger=True, sync_dist=len(self.gpus)>1)
        self.log('train_depth_loss', res['depth_loss'], prog_bar=True, logger=True, sync_dist=len(self.gpus)>1)
        return {'loss': res['depth_loss']}
    
    def validation_step(self, batch, batch_idx, dataset_idx):
        res = self.shared_step(batch, train=False)
        dataset = self.val_datasets[dataset_idx]
        res['dataset'] = dataset
        return res
    
    def register_save_on_error_callback(self, callback):
        '''
            On error, will call the following callback. 
            Callback should have signature:
                callback(batch) -> none
        '''
        self.on_error_callback = callback
        self.save_debug_info_on_error = True
        
    def shared_step(self, batch, train=True):
        try:
            return self._shared_step(batch, train)
        except:
            if self.save_debug_info_on_error:
                self.on_error_callback(batch)
            raise

    def make_valid_mask(self, mask_float, max_pool_size=4, return_small_mask=False):
        '''
            Creates a mask indicating the valid parts of the image(s).
            Enlargens masked area using a max pooling operation.

            Args:
                mask_float: A mask as loaded from the Taskonomy loader.
                max_pool_size: Parameter to choose how much to enlarge masked area.
                return_small_mask: Set to true to return mask for aggregated image
        '''
        if len(mask_float.shape) == 3:
            mask_float = mask_float.unsqueeze(axis=0)
        elif len(mask_float.shape) == 2:
            mask_float = mask_float.unsqueeze(axis=0).unsqueeze(axis=0)

        h, w = mask_float.shape[2], mask_float.shape[3]
        reshape_temp = len(mask_float.shape) == 5
        if reshape_temp:
            mask_float = rearrange(mask_float, 'b p c h w -> (b p) c h w')
        mask_float = 1 - mask_float
        mask_float = F.max_pool2d(mask_float, kernel_size=max_pool_size)
        # mask_float = F.interpolate(mask_float, (self.image_size, self.image_size), mode='nearest')
        mask_float = F.interpolate(mask_float, (h, w), mode='nearest')
        mask_valid = mask_float == 0
        if reshape_temp:
            mask_valid = rearrange(mask_valid, '(b p) c h w -> b p c h w', p=self.num_positive)

        return mask_valid

    
    def _shared_step(self, batch, train=True):
        
        if train:
            # resize augmentation
            batch['positive'] = self.aug.resize_augmentation(batch['positive'], self.tasks, fixed_size=384)
            # rgb augmentation
            augmented_rgb = self.aug.augment_rgb(batch)
        else:
            # rgb augmentation
            augmented_rgb = batch['positive']['rgb']

        step_results = {}
        depth_gt = batch['positive']['depth_zbuffer']
        # augmented_rgb = batch['positive']['rgb']

        # Forward pass
        depth_preds = self(augmented_rgb)
        # clamp the output
        depth_preds = torch.clamp(depth_preds, 0, 1)

        # Mask out invalid pixels and compute loss
        mask_valid = self.make_valid_mask(batch['positive']['mask_valid'])

        # MiDaS Loss
        midas_loss, ssi_loss, reg_loss = self.midas_loss(depth_preds, depth_gt, mask_valid)

        # Virtual Normal Loss
        vn_loss = self.vnl_loss(depth_preds, depth_gt)

        if train and self.global_step < 15000:
            loss = ssi_loss
            reg_loss = 0
            vn_loss = 0
        else:
            loss = ssi_loss + 0.1 * reg_loss + 10 * vn_loss

        step_results.update({
            'ssi_loss': ssi_loss,
            'reg_loss': reg_loss,
            'vn_loss': vn_loss,
            'depth_loss': loss
        })
        return step_results


    def validation_epoch_end(self, outputs):
        loss_counts = 0
        loss = 0
        for dataloader_outputs in outputs:
            for output in dataloader_outputs:
                loss_counts += 1
                loss += output['depth_loss']
 
        loss /= loss_counts
        self.log(f'val_depth_loss', loss, prog_bar=False, logger=True, sync_dist=len(self.gpus)>1)
       
        # Log validation set images 
        if self.global_step >= self.last_log_step + self.log_step or self.global_step<1000:
            self.last_log_step = self.global_step
            self.log_validation_example_images()

    def select_val_samples_for_datasets(self):
        frls = 0
        val_imgs = defaultdict(list)
        for dataset, valset in self.valsets.items():
            while len(val_imgs[dataset]) < self.num_val_images:
                idx = random.randint(0, len(valset) - 1)
                val_imgs[dataset].append(idx)

        return val_imgs
    
    def depth_to_rgb(self, img):   
        img = (img - np.min(img)) / np.ptp(img)
        cm = plt.get_cmap('inferno', 2**16)
        # pixel_colored = np.uint8(np.rint(cm(1-img) * 255))[:, :, :3]
        pixel_colored = np.uint8(img * 255)
        return pixel_colored

    def inverse_normalize(self, y):
        mean = self.normalization_mean
        std = self.normalization_std
        x = y.new(*y.size())
        x[:, 0, :, :] = y[:, 0, :, :] * std[0] + mean[0]
        x[:, 1, :, :] = y[:, 1, :, :] * std[1] + mean[1]
        x[:, 2, :, :] = y[:, 2, :, :] * std[2] + mean[2]
        return x

    def log_validation_example_images(self):

        self.model.eval()
        all_imgs = defaultdict(list)

        for dataset in self.valsets.keys():
            save_path = os.path.join(self.save_dir, 'images', self.experiment_name, dataset)
            if not os.path.exists(save_path):
                os.makedirs(save_path)
            for img_idx in self.val_samples[dataset]:
                example = self.valsets[dataset][img_idx]

                rgb_pos = example['positive']['rgb'].to(self.device)

                depth_gt_pos = example['positive']['depth_zbuffer'] #.squeeze(0)

                mask_valid = self.make_valid_mask(example['positive']['mask_valid'][0]).squeeze(axis=0)
                # mask_valid = self.make_valid_mask(example['positive']['mask_valid']).squeeze(axis=0)

                depth_gt_pos[~mask_valid] = 0

                rgb_pos = rgb_pos.unsqueeze(axis=0)
                depth_gt_pos = depth_gt_pos.unsqueeze(axis=0)

                with torch.no_grad():
                    depth_preds_pos = self.model.forward(rgb_pos).unsqueeze(1) 
                    depth_preds_pos = torch.clamp(depth_preds_pos, 0, 1)
                    depth_preds_pos[~mask_valid.unsqueeze(0)] = 0

                rgb = self.inverse_normalize(rgb_pos)
                rgb = rgb[0].permute(1, 2, 0).detach().cpu().numpy()
                rgb_im = Image.fromarray(np.uint8(255 * rgb))
                rgb_im.save(os.path.join(save_path, f'{img_idx}-rgb.png'))

                mask = mask_valid.permute(1, 2, 0).detach().cpu().numpy()
                depth_gt = depth_gt_pos[0].permute(1, 2, 0).detach().cpu().numpy()
                depth_gt = self.depth_to_rgb(depth_gt)
                depth_gt[~mask] = 0
                depth_gt_im = Image.fromarray(depth_gt.squeeze())
                depth_gt_im.save(os.path.join(save_path, f'{img_idx}-gt-depthzbuffer.png'))

                depth_pred = depth_preds_pos[0].permute(1, 2, 0).detach().cpu().numpy()
                depth_pred = self.depth_to_rgb(depth_pred)
                depth_pred[~mask] = 0
                depth_pred_im = Image.fromarray(depth_pred.squeeze())
                depth_pred_im.save(os.path.join(save_path, f'{img_idx}-pred-depthzbuffer.png'))
                

    
    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        return optimizer



def save_model_and_batch_on_error(checkpoint_function, save_path_prefix='.'):
    def _save(batch):
        checkpoint_function(os.path.join(save_path_prefix, "crash_model.pth"))
        print(f"Saving crash information to {save_path_prefix}")
        with open(os.path.join(save_path_prefix, "crash_batch.pth"), 'wb') as f:
            torch.save(batch, f)
        
    return _save





if __name__ == '__main__':
    # Experimental setup
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--config_file', type=str, default='config/depth.yml',
        help='Path to the config file. (default: config/depth.yml)')
    parser.add_argument(
        '--experiment_name', type=str, default='exp1',
        help='Experiment name for logging and saving checkpoints. (default: exp1)')

    # Add PyTorch Lightning Module and Trainer args
    parser = Trainer.add_argparse_args(parser)
    args = parser.parse_args()

    model = Depth(args.config_file, args.experiment_name)


    # Save best and last model 
    checkpoint_dir = os.path.join(model.save_dir, 'checkpoints', f'{args.experiment_name}')
    checkpoint_callback = ModelCheckpoint(
        filepath=os.path.join(checkpoint_dir, '{epoch}'),
        verbose=True, monitor='val_depth_loss', mode='min', period=1, save_last=True, save_top_k=model.save_top_k
    )

    trainer = Trainer.from_argparse_args(args, checkpoint_callback=checkpoint_callback,  \
                                        gpus=model.gpus, auto_lr_find=False, gradient_clip_val=10,\
                                        accelerator='ddp', replace_sampler_ddp=False)


    model.register_save_on_error_callback(
        save_model_and_batch_on_error(
            trainer.save_checkpoint,
            model.save_dir
        )
    )

    trainer.fit(model)


import torch
import torch.nn as nn
import numpy as np

cuda0 = torch.device('cuda:1')

class VNL_Loss(nn.Module):
    """
    Virtual Normal Loss Function.
    """
    def __init__(self, focal_x, focal_y, input_size,
                 delta_cos=0.867, delta_diff_x=0.01,
                 delta_diff_y=0.01, delta_diff_z=0.01,
                 delta_z=0.0001, sample_ratio=0.15):
        super(VNL_Loss, self).__init__()
        self.fx = torch.tensor([focal_x], dtype=torch.float32) #.to(cuda0)
        self.fy = torch.tensor([focal_y], dtype=torch.float32) #.to(cuda0)
        self.input_size = input_size
        self.u0 = torch.tensor(input_size[1] // 2, dtype=torch.float32) #.to(cuda0)
        self.v0 = torch.tensor(input_size[0] // 2, dtype=torch.float32) #.to(cuda0)
        self.init_image_coor()
        self.delta_cos = delta_cos
        self.delta_diff_x = delta_diff_x
        self.delta_diff_y = delta_diff_y
        self.delta_diff_z = delta_diff_z
        self.delta_z = delta_z
        self.sample_ratio = sample_ratio

    def init_image_coor(self):
        x_row = np.arange(0, self.input_size[1])
        x = np.tile(x_row, (self.input_size[0], 1))
        x = x[np.newaxis, :, :]
        x = x.astype(np.float32)
        x = torch.from_numpy(x.copy()) #.to(cuda0)
        self.u_u0 = x - self.u0

        y_col = np.arange(0, self.input_size[0])  # y_col = np.arange(0, height)
        y = np.tile(y_col, (self.input_size[1], 1)).T
        y = y[np.newaxis, :, :]
        y = y.astype(np.float32)
        y = torch.from_numpy(y.copy()) #.to(cuda0)
        self.v_v0 = y - self.v0

    def transfer_xyz(self, depth):
        # print('!!!!!!!!!!!!!!!111111 ', self.u_u0.device, torch.abs(depth).device, self.fx.device)
        x = self.u_u0 * torch.abs(depth) / self.fx
        y = self.v_v0 * torch.abs(depth) / self.fy
        z = depth
        pw = torch.cat([x, y, z], 1).permute(0, 2, 3, 1) # [b, h, w, c]
        return pw

    def select_index(self):
        valid_width = self.input_size[1]
        valid_height = self.input_size[0]
        num = valid_width * valid_height
        p1 = np.random.choice(num, int(num * self.sample_ratio), replace=True)
        np.random.shuffle(p1)
        p2 = np.random.choice(num, int(num * self.sample_ratio), replace=True)
        np.random.shuffle(p2)
        p3 = np.random.choice(num, int(num * self.sample_ratio), replace=True)
        np.random.shuffle(p3)

        p1_x = p1 % self.input_size[1]
        p1_y = (p1 / self.input_size[1]).astype(np.int)

        p2_x = p2 % self.input_size[1]
        p2_y = (p2 / self.input_size[1]).astype(np.int)

        p3_x = p3 % self.input_size[1]
        p3_y = (p3 / self.input_size[1]).astype(np.int)
        p123 = {'p1_x': p1_x, 'p1_y': p1_y, 'p2_x': p2_x, 'p2_y': p2_y, 'p3_x': p3_x, 'p3_y': p3_y}
        return p123

    def form_pw_groups(self, p123, pw):
        """
        Form 3D points groups, with 3 points in each grouup.
        :param p123: points index
        :param pw: 3D points
        :return:
        """
        p1_x = p123['p1_x']
        p1_y = p123['p1_y']
        p2_x = p123['p2_x']
        p2_y = p123['p2_y']
        p3_x = p123['p3_x']
        p3_y = p123['p3_y']

        pw1 = pw[:, p1_y, p1_x, :]
        pw2 = pw[:, p2_y, p2_x, :]
        pw3 = pw[:, p3_y, p3_x, :]
        # [B, N, 3(x,y,z), 3(p1,p2,p3)]
        pw_groups = torch.cat([pw1[:, :, :, np.newaxis], pw2[:, :, :, np.newaxis], pw3[:, :, :, np.newaxis]], 3)
        return pw_groups

    def filter_mask(self, p123, gt_xyz, delta_cos=0.867,
                    delta_diff_x=0.005,
                    delta_diff_y=0.005,
                    delta_diff_z=0.005):
        pw = self.form_pw_groups(p123, gt_xyz)
        pw12 = pw[:, :, :, 1] - pw[:, :, :, 0]
        pw13 = pw[:, :, :, 2] - pw[:, :, :, 0]
        pw23 = pw[:, :, :, 2] - pw[:, :, :, 1]
        ###ignore linear
        pw_diff = torch.cat([pw12[:, :, :, np.newaxis], pw13[:, :, :, np.newaxis], pw23[:, :, :, np.newaxis]],
                            3)  # [b, n, 3, 3]
        m_batchsize, groups, coords, index = pw_diff.shape
        proj_query = pw_diff.view(m_batchsize * groups, -1, index).permute(0, 2, 1)  # (B* X CX(3)) [bn, 3(p123), 3(xyz)]
        proj_key = pw_diff.view(m_batchsize * groups, -1, index)  # B X  (3)*C [bn, 3(xyz), 3(p123)]
        q_norm = proj_query.norm(2, dim=2)
        nm = torch.bmm(q_norm.view(m_batchsize * groups, index, 1), q_norm.view(m_batchsize * groups, 1, index)) #[]
        energy = torch.bmm(proj_query, proj_key)  # transpose check [bn, 3(p123), 3(p123)]
        norm_energy = energy / (nm + 1e-8)
        norm_energy = norm_energy.view(m_batchsize * groups, -1)
        mask_cos = torch.sum((norm_energy > delta_cos) + (norm_energy < -delta_cos), 1) > 3  # igonre
        mask_cos = mask_cos.view(m_batchsize, groups)
        ##ignore padding and invilid depth
        mask_pad = torch.sum(pw[:, :, 2, :] > self.delta_z, 2) == 3

        ###ignore near
        mask_x = torch.sum(torch.abs(pw_diff[:, :, 0, :]) < delta_diff_x, 2) > 0
        mask_y = torch.sum(torch.abs(pw_diff[:, :, 1, :]) < delta_diff_y, 2) > 0
        mask_z = torch.sum(torch.abs(pw_diff[:, :, 2, :]) < delta_diff_z, 2) > 0

        mask_ignore = (mask_x & mask_y & mask_z) | mask_cos
        mask_near = ~mask_ignore
        mask = mask_pad & mask_near

        return mask, pw

    def select_points_groups(self, gt_depth, pred_depth):
        pw_gt = self.transfer_xyz(gt_depth)
        pw_pred = self.transfer_xyz(pred_depth)
        B, C, H, W = gt_depth.shape
        p123 = self.select_index()
        # mask:[b, n], pw_groups_gt: [b, n, 3(x,y,z), 3(p1,p2,p3)]
        mask, pw_groups_gt = self.filter_mask(p123, pw_gt,
                                              delta_cos=0.867,
                                              delta_diff_x=0.005,
                                              delta_diff_y=0.005,
                                              delta_diff_z=0.005)

        # [b, n, 3, 3]
        pw_groups_pred = self.form_pw_groups(p123, pw_pred)
        pw_groups_pred[pw_groups_pred[:, :, 2, :] == 0] = 0.0001
        mask_broadcast = mask.repeat(1, 9).reshape(B, 3, 3, -1).permute(0, 3, 1, 2)
        pw_groups_pred_not_ignore = pw_groups_pred[mask_broadcast].reshape(1, -1, 3, 3)
        pw_groups_gt_not_ignore = pw_groups_gt[mask_broadcast].reshape(1, -1, 3, 3)

        return pw_groups_gt_not_ignore, pw_groups_pred_not_ignore

    def forward(self, gt_depth, pred_depth, select=True):
        """
        Virtual normal loss.
        :param pred_depth: predicted depth map, [B,W,H,C]
        :param data: target label, ground truth depth, [B, W, H, C], padding region [padding_up, padding_down]
        :return:
        """
        device = gt_depth.device
        self.fx = self.fx.to(device)
        self.fy = self.fy.to(device)
        self.u0 = self.u0.to(device)
        self.v0 = self.v0.to(device)
        self.u_u0 = self.u_u0.to(device)
        self.v_v0 = self.v_v0.to(device)
        # print("************ ", self.fx.device, self.u_u0.device)

        gt_points, dt_points = self.select_points_groups(gt_depth, pred_depth)

        gt_p12 = gt_points[:, :, :, 1] - gt_points[:, :, :, 0]
        gt_p13 = gt_points[:, :, :, 2] - gt_points[:, :, :, 0]
        dt_p12 = dt_points[:, :, :, 1] - dt_points[:, :, :, 0]
        dt_p13 = dt_points[:, :, :, 2] - dt_points[:, :, :, 0]

        gt_normal = torch.cross(gt_p12, gt_p13, dim=2)
        dt_normal = torch.cross(dt_p12, dt_p13, dim=2)
        dt_norm = torch.norm(dt_normal, 2, dim=2, keepdim=True)
        gt_norm = torch.norm(gt_normal, 2, dim=2, keepdim=True)
        dt_mask = dt_norm == 0.0
        gt_mask = gt_norm == 0.0
        dt_mask = dt_mask.to(torch.float32)
        gt_mask = gt_mask.to(torch.float32)
        dt_mask *= 0.01
        gt_mask *= 0.01
        gt_norm = gt_norm + gt_mask
        dt_norm = dt_norm + dt_mask
        gt_normal = gt_normal / gt_norm
        dt_normal = dt_normal / dt_norm
        loss = torch.abs(gt_normal - dt_normal)
        loss = torch.sum(torch.sum(loss, dim=2), dim=0)
        if select:
            loss, indices = torch.sort(loss, dim=0, descending=False)
            loss = loss[int(loss.size(0) * 0.25):]
        loss = torch.mean(loss)
        return loss


# if __name__ == '__main__':
#     import cv2
#     vnl_loss = VNL_Loss(1.0, 1.0, (480, 640))
#     pred_depth = np.ones([2, 1, 480, 640])
#     gt_depth = np.ones([2, 1, 480, 640])
#     gt_depth = torch.tensor(np.asarray(gt_depth, np.float32)).cuda()
#     pred_depth = torch.tensor(np.asarray(pred_depth, np.float32)).cuda()
#     loss = vnl_loss.cal_VNL_loss(pred_depth, gt_depth)
#     print(loss)



from   collections import namedtuple, Counter, defaultdict
from   dataclasses import dataclass, field
import os
import pickle
from   PIL import Image, ImageFile
import re
import torch
import torch.utils.data as data
from   typing import Optional, List, Callable, Union, Dict, Any, Tuple
import warnings

from .masks import make_mask_from_data, DEFAULT_MASK_EXTRA_RADIUS
from .splits import taskonomy_flat_split_to_buildings
from .transforms import default_loader, get_transform
from .task_configs import task_parameters, SINGLE_IMAGE_TASKS #, *

ImageFile.LOAD_TRUNCATED_IMAGES = True # TODO Test this


LabelFile = namedtuple('LabelFile', ['point', 'view', 'domain'])
View = namedtuple('View', ['building', 'point', 'view'])

                   
class TaskonomyDataset(data.Dataset):
    '''
        Loads data for the Taskonomy dataset.
        This expects that the data is structured
        
            /path/to/data/
                rgb/
                    modelk/
                        point_i_view_j.png
                        ...                        
                depth_euclidean/
                ... (other tasks)
                
        If one would like to use pretrained representations, then they can be added into the directory as:
            /path/to/data/
                rgb_encoding/
                    modelk/
                        point_i_view_j.npy
                ...
        
        Basically, any other folder name will work as long as it is named the same way.
    '''
    @dataclass
    class Options():
        '''
            data_path: Path to data. Either one for all, or a dictionary mapping tasks to their data paths.
            tasks: Which tasks to load. Any subfolder will work as long as data is named accordingly
            buildings: Which models to include. See `splits.taskonomy` (can also be a string, e.g. 'fullplus-val')
            transform: one transform per task.
            
            Note: This assumes that all images are present in all (used) subfolders
        '''
        data_path: Union[str, Dict[str, str]] = '/scratch/group/taskonomy'
        tasks: List[str] = field(default_factory=lambda: ['rgb'])
        buildings: Union[List[str], str] = 'tiny'
        transform: Optional[Union[List[Callable], str]] = "DEFAULT"  # List[Transform], None, "DEFAULT"
        load_to_mem: bool = False
        zip_file_name: bool = False
        return_mask: bool = False
        mask_extra_radius: int = DEFAULT_MASK_EXTRA_RADIUS
        image_size: Optional[int]=None
        force_refresh_tmp: bool = True

        
    def __init__(self, options: Options):
        self.return_tuple = True
        if isinstance(options.tasks, str):
            options.tasks = [options.tasks]
            options.transform = [options.transform]            
            self.return_tuple = False
        
        self.buildings = taskonomy_flat_split_to_buildings[options.buildings] if isinstance(options.buildings, str) else options.buildings
        self.cached_data = {}
        self.data_path = options.data_path
        self.image_size = options.image_size
        self.load_to_mem = options.load_to_mem
        self.mask_extra_radius = options.mask_extra_radius
        self.return_mask = options.return_mask
        self.tasks = options.tasks
        self.zip_file_name = options.zip_file_name

        
        self.force_refresh_tmp = options.force_refresh_tmp

        # Load saved image locations if they exist, otherwise create and save them
        tmp_path = './tmp/taskonomy_{}_{}.pkl'.format(
            '-'.join(options.tasks), 
            '-'.join(options.buildings) if isinstance(options.buildings, list) else options.buildings
        )
        tmp_exists = os.path.exists(tmp_path)
        if tmp_exists and not self.force_refresh_tmp:
            with open(tmp_path, 'rb') as f:
                self.urls = pickle.load(f)
            self.size = len(self.urls[self.tasks[0]])
            print(f'Loaded TaskonomyDataset with {self.size} images from tmp.')
        else:
            self.urls = {task: make_dataset(os.path.join(self.data_path, task), self.buildings)
                        for task in options.tasks}
          
            
        
#         self.urls = {task: make_dataset(os.path.join(self.data_path, task), self.buildings)
            self.urls, self.size  = self._remove_unmatched_images()
            
            # Save extracted URLs
            os.makedirs(os.path.dirname(tmp_path), exist_ok=True)
            with open(tmp_path, 'wb') as f:
                pickle.dump(self.urls, f)
        
        self.transform = options.transform
        if isinstance(self.transform, str):
            if self.transform == 'DEFAULT':
                self.transform = [get_transform(task, self.image_size) for task in self.tasks]
            else:
                raise ValueError('TaskonomyDataset option transform must be a List[Callable], None, or "DEFAULT"')
        

        # Perhaps load some things into main memory
        # Note (March 2020: This is likely broken due to the transforms not necessarily being a list)
        if self.load_to_mem: 
            print('Writing activations to memory')
            for t, task in zip(self.transform, tasks):
                self.cached_data[task] = [None] * len(self)
                for i, url in enumerate(self.urls[task]):
                    self.cached_data[task][i] = t(default_loader(url))
                self.cached_data[task] = torch.stack(self.cached_data[task])
#             self.cached_data = torch.stack(self.cached_data)
            print('Finished writing some activations to memory')
            


    def __len__(self):
        return self.size

    def __getitem__(self, index):
        mask, mask_needed = None, self.return_mask
        fpaths = [self.urls[task][index] for task in self.tasks]
        
        if self.load_to_mem:
            result = tuple([self.cached_data[task][index] for task in self.tasks])
        else:
            result = [default_loader(path) for path in fpaths]
            if self.transform is not None:
                # result = [transform(tensor) for transform, tensor in zip(self.transform, result)]
                result_post = []
                for i, (transform, tensor) in enumerate(zip(self.transform, result)):
                    try:
                        if transform is not None:
                            result_post.append(transform(tensor))
                        else:
                            result_post.append(tensor)
                        if mask_needed and 'mask_val' in task_parameters[self.tasks[i]]:
                            mask = make_mask(transform(tensor), self.tasks[i])
                            mask_needed = False
                    except Exception as e:
                        print('exception in task:', self.tasks[i], transform, tensor)
                        raise e
                result = result_post

        # handle 2 channel outputs
        for i in range(len(self.tasks)):
            task = self.tasks[i]
            base_task = [t for t in SINGLE_IMAGE_TASKS if t == task]
            if len(base_task) == 0:
                continue
            else:
                base_task = base_task[0]
            num_channels = task_parameters[base_task]['out_channels']
            if 'decoding' in task and result[i].shape[0] != num_channels:
                assert torch.sum(result[i][num_channels:,:,:]) < 1e-5, 'unused channels should be 0.'
                result[i] = result[i][:num_channels,:,:]


        
        result_dict = {task: value for task, value in zip(self.tasks, result)}

        if self.return_mask:
            if mask_needed:
                raise NotImplementedError('Mask not created!')
            result_dict['mask'] = mask
    
        if self.zip_file_name:
            for fpath, task in zip(fpaths, self.tasks):
                result_dict[task + '_fpath'] = fpath 

        if self.return_tuple:
            return result_dict
        else:
            return result[0]
         
    
    def make_mask(self, data: Dict[str, Any]):
        ''' data: Map of tasks -> Tensors '''
        tasks = sorted(data.keys())
        return make_mask_from_data(
                            [data[t] for t in tasks],
                            tasks,
                            self.mask_extra_radius)
    
    def task_config(self, task):
        return task_parameters[task]


    
    def _remove_unmatched_images(self) -> (Dict[str, List[str]], int):
        '''
            Filters out point/view/building triplets that are not present for all tasks
            
            Returns:
                filtered_urls: Filtered Dict
                max_length: max([len(urls) for _, urls in filtered_urls.items()])
        '''
        n_images_task = [(len(obs), task) for task, obs in self.urls.items()]
        max_images = max(n_images_task)[0]
        if max(n_images_task)[0] == min(n_images_task)[0]:
            return self.urls, max_images
        else:
            print("Each task must have the same number of images. However, the max != min ({} != {}). Number of images per task is: \n\t{}".format(
                max(n_images_task)[0], min(n_images_task)[0], "\n\t".join([str(t) for t in n_images_task])))
            # Get views for each task
            
            
            def _parse_fpath_for_view( path ):
                building = os.path.basename(os.path.dirname(path))
                file_name = os.path.basename(path) 
                lf = parse_filename( file_name )
                return View(view=lf.view, point=lf.point, building=building)

            self.task_to_view = {}
            for task, paths in self.urls.items():
                self.task_to_view[task] = [_parse_fpath_for_view( path ) for path in paths]
            
            # Compute intersection
            intersection = None
            for task, uuids in self.task_to_view.items():
                if intersection is None:
                    intersection = set(uuids)
                else:
                    intersection = intersection.intersection(uuids)

#             print(sorted(list(intersection)))
            # Keep intersection
            print('Keeping intersection: ({} images/task)...'.format(len(intersection)))
            new_urls = {}
            for task, paths in self.urls.items():
                new_urls[task] = [path for path in paths if _parse_fpath_for_view( path ) in intersection]
#                 for path in paths:
#                     building = os.path.basename(os.path.dirname(path))
#                     file_name = os.path.basename(path) 
#                     lf = parse_filename( file_name )
#                     view = str(View(view=lf.view, point=lf.point, building=building))
#                     print(view)
#                     if view in intersection:
#                         paths_for_task.append(path)
#                 new_urls[task] = paths_for_task
            return new_urls, len(intersection)
        raise NotImplementedError('Reached the end of this function. You should not be seeing this!')

    
    def _validate_images_per_building(self):
            # Validate number of images
            print("Building TaskonomyDataset:")
            n_images_task = [(len(obs), task) for task, obs in self.urls.items()]
            print("\t" + "  |  ".join(["{}: {}".format(k, task) for task, k in n_images_task]))
            if max(n_images_task)[0] != min(n_images_task)[0]:
                print("Each task must have the same number of images. However, the max != min ({} != {}). Number of images per task is: \n\t{}".format(
                    max(n_images_task)[0], min(n_images_task)[0], "\n\t".join([str(t) for t in n_images_task])))

                # count number of frames per building per task
                all_building_counts = defaultdict(dict)
                for task, obs in self.urls.items():
                    c = Counter([url.split("/")[-2] for url in obs])
                    for building in c:
                        all_building_counts[building][task] = c[building]

                # find where the number of distinct counts is more than 1
                print('Removing data from the following buildings')
                buildings_to_remove = []
                for b, count in all_building_counts.items():
                    if len(set(list(count.values()))) > 1:
                        print(f"\t{b}:", count)
                        buildings_to_remove.append(b)
                    if len(count) != len(self.tasks):
                        print(f"\t{b}: missing in tasks", set(self.tasks) - set(count.keys()))
                        buildings_to_remove.append(b)
                # [(len(obs), task) for task, obs in self.urls.items()]

                # redo the loading with fewer buildings
                buildings_redo = [b for b in self.buildings if b not in buildings_to_remove]
                self.urls = {task: make_dataset(os.path.join(self.data_path, task), buildings_redo)
                            for task in self.tasks}
                n_images_task = [(len(obs), task) for task, obs in self.urls.items()]
                print("\t" + "  |  ".join(["{}: {}".format(k, task) for task, k in n_images_task]))
            assert max(n_images_task)[0] == min(n_images_task)[0], \
                    "Each task must have the same number of images. However, the max != min ({} != {}). Number of images per task is: \n\t{}".format(
                    max(n_images_task)[0], min(n_images_task)[0], "\n\t".join([str(t) for t in n_images_task]))
            return n_images_task


def make_dataset(dir, folders=None):
    #  folders are building names. If None, get all the images (from both building folders and dir)
    images = []
    dir = os.path.expanduser(dir)
    if not os.path.isdir(dir):
        assert "bad directory"

    for subfolder in sorted(os.listdir(dir)):
        subfolder_path = os.path.join(dir, subfolder)
        if os.path.isdir(subfolder_path) and (folders is None or subfolder in folders):
            for fname in sorted(os.listdir(subfolder_path)):
                path = os.path.join(subfolder_path, fname)
                images.append(path)

        # If folders/buildings are not specified, use images in dir
        if folders is None and os.path.isfile(subfolder_path):
            images.append(subfolder_path)

    return images



def parse_filename( filename ):
    p = re.match('.*point_(?P<point>\d+)_view_(?P<view>\d+)_domain_(?P<domain>\w+)', filename)
    if p is None:
        raise ValueError( 'Filename "{}" not matched. Must be of form point_XX_view_YY_domain_ZZ.**.'.format(filename) )

    lf = {'point': p.group('point'), 'view': p.group('view'), 'domain': p.group('domain') }
    return LabelFile(**lf)


class TaskonomyDataLoader:

    @dataclass
    class Options(TaskonomyDataset.Options):
        phase: str = 'val'
        batch_size: int = 6
        shuffle: bool = True
        num_workers: int = 8
        pin_memory: bool = True

    def make(options: Options):
        is_train = (options.phase == 'train')
        dataset = TaskonomyDataset(options)
        return data.DataLoader(
            dataset=dataset,
            batch_size=options.batch_size,
            shuffle=options.shuffle,
            num_workers=options.num_workers,
            pin_memory=options.pin_memory,
            drop_last=is_train)



import torch
import torch.nn as nn
import torch.nn.functional as F

from .base_model import BaseModel
from .blocks import (
    FeatureFusionBlock,
    FeatureFusionBlock_custom,
    Interpolate,
    _make_encoder,
    forward_vit,
)


def _make_fusion_block(features, use_bn):
    return FeatureFusionBlock_custom(
        features,
        nn.ReLU(False),
        deconv=False,
        bn=use_bn,
        expand=False,
        align_corners=True,
    )


class DPT(BaseModel):
    def __init__(
        self,
        head,
        features=256,
        backbone="vitb_rn50_384",
        readout="project",
        channels_last=False,
        use_bn=False,
    ):

        super(DPT, self).__init__()

        self.channels_last = channels_last

        hooks = {
            "vitb_rn50_384": [0, 1, 8, 11],
            "vitb16_384": [2, 5, 8, 11],
            "vitl16_384": [5, 11, 17, 23],
        }

        # Instantiate backbone and reassemble blocks
        self.pretrained, self.scratch = _make_encoder(
            backbone,
            features,
            True, # Set to true of you want to train from scratch, uses ImageNet weights
            groups=1,
            expand=False,
            exportable=False,
            hooks=hooks[backbone],
            use_readout=readout,
        )

        self.scratch.refinenet1 = _make_fusion_block(features, use_bn)
        self.scratch.refinenet2 = _make_fusion_block(features, use_bn)
        self.scratch.refinenet3 = _make_fusion_block(features, use_bn)
        self.scratch.refinenet4 = _make_fusion_block(features, use_bn)

        self.scratch.output_conv = head


    def forward(self, x):
        if self.channels_last == True:
            x.contiguous(memory_format=torch.channels_last)

        layer_1, layer_2, layer_3, layer_4 = forward_vit(self.pretrained, x)

        layer_1_rn = self.scratch.layer1_rn(layer_1)
        layer_2_rn = self.scratch.layer2_rn(layer_2)
        layer_3_rn = self.scratch.layer3_rn(layer_3)
        layer_4_rn = self.scratch.layer4_rn(layer_4)

        path_4 = self.scratch.refinenet4(layer_4_rn)
        path_3 = self.scratch.refinenet3(path_4, layer_3_rn)
        path_2 = self.scratch.refinenet2(path_3, layer_2_rn)
        path_1 = self.scratch.refinenet1(path_2, layer_1_rn)

        out = self.scratch.output_conv(path_1)

        return out

class DPTDepthModel(DPT):
    def __init__(self, path=None, non_negative=True, num_channels=1, **kwargs):
        features = kwargs["features"] if "features" in kwargs else 256

        head = nn.Sequential(
            nn.Conv2d(features, features // 2, kernel_size=3, stride=1, padding=1),
            Interpolate(scale_factor=2, mode="bilinear", align_corners=True),
            nn.Conv2d(features // 2, 32, kernel_size=3, stride=1, padding=1),
            nn.ReLU(True),
            nn.Conv2d(32, num_channels, kernel_size=1, stride=1, padding=0),
            nn.ReLU(True) if non_negative else nn.Identity(),
            nn.Identity(),
        )

        super().__init__(head, **kwargs)

        if path is not None:
           self.load(path)

    def forward(self, x):
        return super().forward(x).squeeze(dim=1)


from   collections import namedtuple, Counter, defaultdict
from   dataclasses import dataclass, field
from   joblib import Parallel, delayed
import logging
import multiprocessing as mp
import os
import pickle
import json
from   PIL import Image, ImageFile
import pandas as pd
import random
import re
from   time import perf_counter 
from   tqdm import tqdm
import torch
import torch.utils.data as data
import torch.nn.functional as F
from   torchvision import transforms
import torchvision.transforms.functional as TF
from   typing import Optional, List, Callable, Union, Dict, Any
import warnings

from .taskonomy_dataset import parse_filename, LabelFile, View
from .splits import taskonomy_flat_split_to_buildings, replica_flat_split_to_buildings, \
    gso_flat_split_to_buildings, hypersim_flat_split_to_buildings, blendedMVS_flat_split_to_buildings
from .transforms import default_loader, get_transform, LocalContrastNormalization
from .task_configs import task_parameters, SINGLE_IMAGE_TASKS
from .refocus_augmentation import RefocusImageAugmentation

ImageFile.LOAD_TRUNCATED_IMAGES = True # TODO Test this

MAX_VIEWS = 45

RGB_MEAN = torch.Tensor([0.55312, 0.52514, 0.49313]).reshape(3,1,1)
RGB_STD =  torch.Tensor([0.20555, 0.21775, 0.24044]).reshape(3,1,1)

REPLICA_BUILDINGS = [
    'frl_apartment_5', 'office_2', 'room_2', 'office_4', 'frl_apartment_0', 'frl_apartment_4',
    'office_1', 'frl_apartment_3', 'office_0', 'apartment_2', 'room_0', 'apartment_1', 
    'frl_apartment_1', 'office_3', 'frl_apartment_2', 'apartment_0', 'hotel_0', 'room_1']

def filter_taskonomy():
    data = pd.read_csv('/scratch/roman/Omnidata/loss_stats/taskonomy_sorted.csv', index_col=0)
    thresholds = [10000, 50000, 100000]
    filtered_bpv = {}
    for thresh in thresholds:
        count = 0
        d = defaultdict(lambda: defaultdict(list))
        for index, row in data.iterrows():
            if count > thresh: break
            d[row['building']][int(row['point'])].append(int(row['view']))
            count += 1
        filtered_bpv[thresh] = d
    return filtered_bpv

def filter_blendedMVS():
    data = pd.read_csv('/scratch/roman/Omnidata/loss_stats/blendedMVS_sorted.csv', index_col=0)
    thresholds = [1000, 10000, 20000]
    filtered_bpv = {}
    for thresh in thresholds:
        count = 0
        d = defaultdict(list)
        for index, row in data.iterrows():
            if count > thresh: break
            d[row['building']].append(int(row['point']))
            count += 1
        filtered_bpv[thresh] = d
    return filtered_bpv

blendedMVS_bad_scenes = ['5bf21799d43923194842c001',
                        '000000000000000000000001',
                        '000000000000000000000000',
                        '5b21e18c58e2823a67a10dd8',
                        '5ab8b8e029f5351f7f2ccf59',
                        '00000000000000000000000c',
                        '00000000000000000000000b',
                        '58cf4771d0f5fb221defe6da',
                        '5c2b3ed5e611832e8aed46bf',
                        '584a7333fe3cb463906c9fe6',
                        '585ee0632a57cc11d4933608',
                        '5983012d1bd4b175e70c985a',
                        '58a0dd1a3d0b4542479a28f3',
                        '58a44463156b87103d3ed45e']

                    
class OmnidataDataset(data.Dataset):
    '''
        Loads data for the Taskonomy dataset.
        This expects that the data is structured
        
            /path/to/data/
                rgb/
                    modelk/
                        point_i_view_j.png
                        ...                        
                depth_euclidean/
                ... (other tasks)
                
        If one would like to use pretrained representations, then they can be added into the directory as:
            /path/to/data/
                rgb_encoding/
                    modelk/
                        point_i_view_j.npy
                ...
        
        Basically, any other folder name will work as long as it is named the same way.
    '''
    @dataclass
    class Options():
        '''
            data_path: Path to data
            tasks: Which tasks to load. Any subfolder will work as long as data is named accordingly
            buildings: Which models to include. See `splits.taskonomy` (can also be a string, e.g. 'fullplus-val')
            transform: one transform per task.
            
            Note: This assumes that all images are present in all (used) subfolders
        '''
        taskonomy_data_path: str = '/datasets/taskonomy'
        replica_data_path: str = '/scratch/ainaz/replica-taskonomized'
        gso_data_path: str = '/scratch/ainaz/replica-google-objects'
        hypersim_data_path: str = '/scratch/ainaz/hypersim-dataset2/evermotion/scenes'
        blendedMVS_data_path: str = '/scratch/ainaz/BlendedMVS/mvs_low_res_taskonomized'
        split: str = 'train'
        taskonomy_variant: str = 'tiny'
        tasks: List[str] = field(default_factory=lambda: ['rgb'])
        datasets: List[str] = field(default_factory=lambda: ['taskonomy', 'replica', 'gso'])
        transform: Optional[Union[Dict[str, Callable], str]] = "DEFAULT"  # List[Transform], None, "DEFAULT"
        image_size: Optional[int] = None
        num_positive: Union[int, str] = 1 # Either int or 'all'
        normalize_rgb: bool = False
        normalization_mean: List[float] = field(default_factory=lambda: [0.5,0.5,0.5])
        normalization_std: List[float] = field(default_factory=lambda: [0.5,0.5,0.5])
        force_refresh_tmp: bool = False


    def load_datasets(self, options):
        # Load saved image locations if they exist, otherwise create and save them
        self.urls = defaultdict(list)
        self.size = 0

        for dataset in self.datasets:
            
            if dataset == 'taskonomy':
                tmp_path = './tmp/{}_{}_{}.pkl'.format(
                    dataset,
                    '-'.join(options.tasks), 
                    f'{options.taskonomy_variant}-{options.split}'
                )
            else:
                tmp_path = './tmp/{}_{}_{}.pkl'.format(
                    dataset,
                    '-'.join(options.tasks), 
                    options.split
                )

            tmp_exists = os.path.exists(tmp_path)
            print("data temp path : ", tmp_path)

            if tmp_exists and not self.force_refresh_tmp:
                with open(tmp_path, 'rb') as f:
                    dataset_urls = pickle.load(f)
                    for task, urls in dataset_urls.items():
                        if task not in options.tasks: continue
                        # TODO remove later
                        task2 = 'segment_semantic' if task == 'segment_panoptic' else task
                        self.urls[task2] += urls
                dataset_size = len(dataset_urls[self.tasks[0]])
                self.size += dataset_size
                print(f'Loaded {dataset} with {dataset_size} images from tmp.')

            else:
                
                if dataset == 'taskonomy':
                    dataset_urls = {task: make_taskonomy_dataset(
                                            os.path.join(self.taskonomy_data_path, task), task, self.taskonomy_buildings)
                                        for task in options.tasks}
                elif dataset == 'replica':
                    dataset_urls = {task: make_replica_gso_dataset(
                                            self.replica_data_path, task, self.replica_buildings)
                                        for task in options.tasks}
                elif dataset == 'gso':
                    dataset_urls = {task: make_replica_gso_dataset(
                                             self.gso_data_path, task, self.gso_buildings) 
                                        for task in options.tasks}
                elif dataset == 'hypersim':
                    # dataset_urls = {task: make_hypersim_dataset(
                    #                          self.hypersim_data_path, task, self.hypersim_buildings) 
                    #                     for task in options.tasks}
                    dataset_urls = {task: make_hypersim_dataset_orig_split(
                                             self.hypersim_data_path, task, self.split) 
                                        for task in options.tasks}

                elif dataset == 'blendedMVS':
                    dataset_urls = {task: make_blendedMVS_dataset(
                                             self.blendedMVS_data_path, task, self.blendedMVS_buildings) 
                                        for task in options.tasks}

                dataset_urls, dataset_size  = self._remove_unmatched_images(dataset_urls)

                for task, urls in dataset_urls.items():
                    self.urls[task] += urls
                self.size += dataset_size

                # Save extracted URLs
                os.makedirs(os.path.dirname(tmp_path), exist_ok=True)
                with open(tmp_path, 'wb') as f:
                    pickle.dump(dataset_urls, f)

        
    def __init__(self, options: Options):
        start_time = perf_counter()
        
        if isinstance(options.tasks, str):
            options.tasks = [options.tasks]
            options.transform = {options.tasks: options.transform}        
        
        self.taskonomy_data_path = options.taskonomy_data_path
        self.replica_data_path = options.replica_data_path
        self.gso_data_path = options.gso_data_path
        self.hypersim_data_path = options.hypersim_data_path
        self.blendedMVS_data_path = options.blendedMVS_data_path
        self.datasets = options.datasets
        self.split = options.split
        self.image_size = options.image_size
        self.tasks = options.tasks
        self.num_positive = MAX_VIEWS if options.num_positive == 'all' else options.num_positive
        self.normalize_rgb = options.normalize_rgb
        self.normalization_mean = options.normalization_mean
        self.normalization_std = options.normalization_std
        self.force_refresh_tmp = options.force_refresh_tmp

        self.taskonomy_buildings = taskonomy_flat_split_to_buildings[f'{options.taskonomy_variant}-{self.split}']
        self.replica_buildings = replica_flat_split_to_buildings[self.split]
        self.gso_buildings = gso_flat_split_to_buildings[self.split]
        self.hypersim_buildings = hypersim_flat_split_to_buildings[self.split]
        self.blendedMVS_buildings = blendedMVS_flat_split_to_buildings[self.split]

        # ### taskonomy filtering
        # taskonomy_filtered_bpv = filter_taskonomy()
        # taskonomy_train_filter_thresh = 50000
        # taskonomy_val_filter_thresh = 100000
        # self.taskonomy_train_filter = taskonomy_filtered_bpv[taskonomy_train_filter_thresh]
        # self.taskonomy_val_filter = taskonomy_filtered_bpv[taskonomy_val_filter_thresh]


        # ### blendedMVS filtering
        # blendedMVS_filtered_bpv = filter_blendedMVS()
        # blendedMVS_train_filter_thresh = 10000
        # blendedMVS_val_filter_thresh = 20000
        # self.blendedMVS_train_filter = blendedMVS_filtered_bpv[blendedMVS_train_filter_thresh]
        # self.blendedMVS_val_filter = blendedMVS_filtered_bpv[blendedMVS_val_filter_thresh]

        self.load_datasets(options)
        self.transform = options.transform
        if isinstance(self.transform, str):
            if self.transform == 'DEFAULT':
                self.transform = {task: get_transform(task, image_size=None) for task in self.tasks}
            else:
                raise ValueError('TaskonomyDataset option transform must be a Dict[str, Callable], None, or "DEFAULT"')
                
        if self.normalize_rgb and 'rgb' in self.transform:
            self.transform['rgb'] = transforms.Compose(
                self.transform['rgb'].transforms +
                [transforms.Normalize(mean=self.normalization_mean, std=self.normalization_std)]

            )
    
        # Saving some lists and dictionaries for fast lookup

        self.tbpv_dict = {} # Save task -> building -> point -> view dict
        self.url_dict = {}  # Save (task, building, point, view) -> URL dict
        self.bpv_count = {} # Dictionary to check if all (building, point, view) tuples have all tasks
        
        for task in self.tasks:
            self.tbpv_dict[task] = {}
            filter_count = 0
            for url in self.urls[task]:
                if url.__contains__('replica-taskonomized'):
                    building = url.split('/')[-3]
                elif url.__contains__('replica-google-objects'): # e.g apartment_0-3, apartment_0-6
                    building = url.split('/')[-4] + '-' + url.split('/')[-3]
                elif url.__contains__('hypersim'): # e.g ai_001_001-cam_00, ai_001_001-cam_01
                    building = url.split('/')[-5] + '-' + url.split('/')[-3]
                elif url.__contains__('taskonomy'):
                    building = url.split('/')[-2]
                elif url.__contains__('BlendedMVS'):
                    building = url.split('/')[-3]
                else:
                    raise NotImplementedError('Dataset path (url) not recognized!')
                
                file_name = url.split('/')[-1].split('_')
  
                point, view = file_name[1], file_name[3]

                ##################
                # filter Taskonomy
                ##################
                # if building == 'grace' or building == 'rough': continue
                # if self.split == 'train':
                #     if building in self.taskonomy_train_filter.keys() and int(point) in self.taskonomy_train_filter[building].keys() and int(view) in self.taskonomy_train_filter[building][int(point)]: 
                #         filter_count += 1
                #         continue
                # elif self.split in ['val', 'test']:
                #     if building in self.taskonomy_val_filter.keys() and int(point) in self.taskonomy_val_filter[building].keys() and int(view) in self.taskonomy_val_filter[building][int(point)]: 
                #         filter_count += 1
                #         continue

                ##################
                # filter BlendedMVS
                ##################                
                if building in blendedMVS_bad_scenes: continue
                # if self.split == 'train':
                #     if building in self.blendedMVS_train_filter.keys() and int(point) in self.blendedMVS_train_filter[building]: 
                #         filter_count += 1
                #         continue
                # elif self.split in ['val', 'test']:
                #     if building in self.blendedMVS_val_filter.keys() and int(point) in self.blendedMVS_val_filter[building]: 
                #         filter_count += 1
                #         continue
                    

                # Populate url_dict
                self.url_dict[(task, building, point, view)] = url

                # Populate tbpv_dict
                if building not in self.tbpv_dict[task]:
                    self.tbpv_dict[task][building] = {}
                if point not in self.tbpv_dict[task][building]:
                    self.tbpv_dict[task][building][point] = []
                self.tbpv_dict[task][building][point].append(view)

                # Populate bpv_count
                if (building, point, view) not in self.bpv_count:
                    self.bpv_count[(building, point, view)] = 1
                else:
                    self.bpv_count[(building, point, view)] += 1
            
            # print(f"Task : {task} - split : {self.split} - filtered {filter_count} images!")


        # Remove entries that don't have all tasks and create list of all (building, point, view) tuples that contain all tasks
        self.bpv_list = [bpv_tuple for bpv_tuple, count in self.bpv_count.items() if count == len(self.tasks)]

        self.views = {}    # Build dictionary that contains all the views from a certain (building, point) tuple
        self.bpv_dict = {} # Save building -> point -> view dict
        for building, point, view in self.bpv_list:
            # Populate views
            if (building, point) not in self.views:
                self.views[(building, point)] = []
            self.views[(building, point)].append(view)

            # Populate bpv_dict
            if building not in self.bpv_dict:
                self.bpv_dict[building] = {}
            if point not in self.bpv_dict[building]:
                self.bpv_dict[building][point] = []
            self.bpv_dict[building][point].append(view)

            
        random.shuffle(self.bpv_list)
        
        end_time = perf_counter()
        self.num_points = len(self.views)
        self.num_images = len(self.bpv_list)
        self.num_buildings = len(self.bpv_dict)
        
        logger = logging.getLogger(__name__)
        logger.warning("Loaded {} images in {:0.2f} seconds".format(self.num_images, end_time - start_time))
        logger.warning("\t ({} buildings) ({} points) ({} images) for domains {}".format(self.num_buildings, self.num_points, self.num_images, self.tasks))


    def __len__(self):
        return len(self.bpv_list)

    def __getitem__(self, index):
        flip = random.random() > 0.5 
        result = {}
        
        # Anchor building / point / view
        building, point, view = self.bpv_list[index]
        
        positive_views = [view]
        positive_samples = {}   

        for task_num, task in enumerate(self.tasks):
            resize_method = Image.BILINEAR if task in ['rgb'] else Image.NEAREST
            task_samples = []
            for v in positive_views:
                path = self.url_dict[(task, building, point, v)]
                res = default_loader(path)

                if self.transform is not None and self.transform[task] is not None:

                    if path.__contains__('hypersim') or path.__contains__('BlendedMVS'):
                        resize_transform = transforms.Resize(self.image_size, resize_method)
                        res = resize_transform(res)
                        if task_num == 0: 
                            i, j, h, w = transforms.RandomCrop.get_params(
                                res, output_size=(self.image_size, self.image_size))
                        res = TF.crop(res, i, j, h, w)
                        res = self.transform[task](res)

                    else:
                        transform = transforms.Compose([
                            transforms.Resize(self.image_size, resize_method),
                            transforms.CenterCrop(self.image_size),
                            self.transform[task]])
                        res = transform(res)
                    

                # flip augmentation
                if flip: 
                    res = torch.flip(res, [2])
                    if task == 'normal': res[0,:,:] = 1 - res[0,:,:]

                task_samples.append(res)

            task_samples = torch.stack(task_samples) if self.num_positive > 1 else task_samples[0]

            positive_samples[task] = task_samples
        
        positive_samples['point'] = point
        positive_samples['building'] = building
        result['positive'] = positive_samples
        
        
        return result
            


    def randomize_order(self, seed=0):
        random.seed(0)
        random.shuffle(self.bpv_list)
    
    def task_config(self, task):
        return task_parameters[task]

    def _remove_unmatched_images(self, dataset_urls) -> (Dict[str, List[str]], int):
        '''
            Filters out point/view/building triplets that are not present for all tasks
            
            Returns:
                filtered_urls: Filtered Dict
                max_length: max([len(urls) for _, urls in filtered_urls.items()])
        '''
        n_images_task = [(len(obs), task) for task, obs in dataset_urls.items()]
        max_images = max(n_images_task)[0]
        if max(n_images_task)[0] == min(n_images_task)[0]:
            return dataset_urls, max_images
        else:
            print("Each task must have the same number of images. However, the max != min ({} != {}). Number of images per task is: \n\t{}".format(
                max(n_images_task)[0], min(n_images_task)[0], "\n\t".join([str(t) for t in n_images_task])))
            # Get views for each task
            def _parse_fpath_for_view( path ):
                url = path
                if url.__contains__('replica-taskonomized'):
                    building = url.split('/')[-3]
                elif url.__contains__('replica-google-objects'): # e.g apartment_0-3, apartment_0-6
                    building = url.split('/')[-4] + '-' + url.split('/')[-3]
                elif url.__contains__('hypersim'): # e.g ai_001_001-cam_00, ai_001_001-cam_01
                    building = url.split('/')[-5] + '-' + url.split('/')[-3]
                elif url.__contains__('taskonomy'):
                    building = url.split('/')[-2]
                elif url.__contains__('BlendedMVS'):
                    building = url.split('/')[-3]
                # building = os.path.basename(os.path.dirname(path))
                file_name = os.path.basename(path) 
                lf = parse_filename( file_name )
                return View(view=lf.view, point=lf.point, building=building)

            self.task_to_view = {}
            for task, paths in dataset_urls.items():
                self.task_to_view[task] = [_parse_fpath_for_view( path ) for path in paths]
    
            # Compute intersection
            intersection = None
            for task, uuids in self.task_to_view.items():
                if intersection is None:
                    intersection = set(uuids)
                else:
                    intersection = intersection.intersection(uuids)
            # Keep intersection
            print('Keeping intersection: ({} images/task)...'.format(len(intersection)))
            new_urls = {}
            for task, paths in dataset_urls.items():
                new_urls[task] = [path for path in paths if _parse_fpath_for_view( path ) in intersection]
            return new_urls, len(intersection)
        raise NotImplementedError('Reached the end of this function. You should not be seeing this!')



def make_taskonomy_dataset(dir, task, folders=None):
    # TODO remove later
    if task == 'segment_semantic': dir = os.path.join(dir, '..', 'segment_panoptic')
    #  folders are building names. If None, get all the images (from both building folders and dir)
    images = []
    dir = os.path.expanduser(dir)
    if not os.path.isdir(dir):
        assert "bad directory"

    for subfolder in folders:
        subfolder_path = os.path.join(dir, subfolder)
        print(subfolder_path)
        if os.path.isdir(subfolder_path) and (folders is None or subfolder in folders):
            for fname in sorted(os.listdir(subfolder_path)):
                path = os.path.join(subfolder_path, fname)
                images.append(path)

        # If folders/buildings are not specified, use images in dir
        if folders is None and os.path.isfile(subfolder_path):
            images.append(subfolder_path)

    return images

def make_replica_gso_dataset(dir, task, folders=None):
    if task == 'segment_semantic': task = 'semantic'
    #  folders are building names. 
    images = []
    dir = os.path.expanduser(dir)
    if not os.path.isdir(dir):
        assert "bad directory"
    
    if folders is None:
        folders = os.listdir(dir)

    for folder in folders:
        if folder not in REPLICA_BUILDINGS: # gso dataset e.g. apartment_0-3, apartment_0-6, apartment_0-15
            folder_path = os.path.join(dir, folder.split('-')[0], folder.split('-')[1], task)
        else: # replica dataset
            folder_path = os.path.join(dir, folder, task)
        for fname in sorted(os.listdir(folder_path)):
            path = os.path.join(folder_path, fname)
            images.append(path)

    return images

def make_hypersim_dataset(dir, task, folders=None):
    if task == 'segment_semantic': task = 'semantic_hdf5'
    #  folders are building names. 
    images = []
    dir = os.path.expanduser(dir)
    if not os.path.isdir(dir):
        assert "bad directory"

    for folder in folders:
        taskonomized_path = os.path.join(dir, folder, 'taskonomized')
        for camera in os.listdir(taskonomized_path):
            if not camera.startswith('cam'):
                continue
            # filter out bad points from filtered_points.json
            with open(os.path.join(taskonomized_path, camera, 'filtered_points.json')) as json_file:
                bad_points = json.load(json_file)
            folder_path = os.path.join(taskonomized_path, camera, task)
            for fname in sorted(os.listdir(folder_path)):
                point = fname.split('_')[1]
                if point not in bad_points:
                    path = os.path.join(folder_path, fname)
                    images.append(path)

    return images

def make_hypersim_dataset_orig_split(dir, task, split):
    hypersim_orig_split_file = os.path.join(os.path.dirname(__file__), 'splits', f'{split}_hypersim_orig.csv')
    df = pd.read_csv(hypersim_orig_split_file)

    if task == 'segment_semantic': task = 'semantic_hdf5'
    if task in ['normal', 'mask_valid', 'depth_zbuffer']: task = task + '2'
    #  folders are building names. 
    images = []
    dir = os.path.expanduser(dir)
    if not os.path.isdir(dir):
        assert "bad directory"

    folders = [scene for scene in set(df['scene_name'].tolist()) if scene in os.listdir(dir)]
    for folder in folders:
        scene = df.loc[df['scene_name']==folder]
        taskonomized_path = os.path.join(dir, folder, 'taskonomized')
        for camera in os.listdir(taskonomized_path):
            if not camera.startswith('cam'):
                continue
            scene_cam = scene.loc[scene['camera_name']==camera]
            # filter out bad points from filtered_points.json
            # with open(os.path.join(taskonomized_path, camera, 'filtered_points.json')) as json_file:
            #     bad_points = json.load(json_file)
            folder_path = os.path.join(taskonomized_path, camera, task)
            if not os.path.exists(folder_path): continue
            for fname in sorted(os.listdir(folder_path)):
                point = fname.split('_')[1]
                # if point in bad_points: continue
                try:
                    row = scene_cam.loc[scene_cam['frame_id']==int(point)]
                except:
                    continue
                if not row.empty and row.iloc[0]['included_in_public_release'] and row.iloc[0]['split_partition_name'] == split:
                    path = os.path.join(folder_path, fname)
                    images.append(path)

        print(len(images))

    return images

def make_blendedMVS_dataset(dir, task, folders=None):
    if task == 'segment_semantic': task = 'semantic'
    if task == 'mask_valid': task = 'mask_valid_depth'
    #  folders are building names. 
    images = []
    dir = os.path.expanduser(dir)
    if not os.path.isdir(dir):
        assert "bad directory"
    
    if folders is None:
        folders = os.listdir(dir)

    for folder in folders:
        folder_path = os.path.join(dir, folder, task)
        for fname in sorted(os.listdir(folder_path)):
            path = os.path.join(folder_path, fname)
            images.append(path)

    return images


def make_empty_like(data_dict):
    if not isinstance(data_dict, dict):
        if isinstance(data_dict, torch.Tensor):
            return torch.zeros_like(data_dict)
        elif isinstance(data_dict, list):
            return [make_empty_like(d) for d in data_dict]
        else:
            return type(data_dict)()
        raise NotImplementedError

    result = {}
    for k, v in data_dict.items():
        result[k] = make_empty_like(v)
    return result




import aria2p
import multiprocess as mp
from functools import partial
import atexit, os, signal, shutil, tempfile, time
from   subprocess import call, run, Popen, check_output
import tarfile
from tarfile import ReadError
from   typing import Dict, Optional, Iterable
from   argparse import SUPPRESS, ArgumentError

from .metadata import RemoteStorageMetadata, ZippedModel, bcolors, notice, header, license, failure, print_and_log_failure
from .starter_dataset import STARTER_DATASET_REMOTE_SERVER_METADATAS, STARTER_DATA_COMPONENT_TO_SPLIT, STARTER_DATA_COMPONENT_TO_SUBSET, STARTER_DATA_COMPONENTS, STARTER_DATA_LICENSES
from fastcore.script import *
import tqdm
import re
import requests
import urllib

__all__ = ['main']
FORBIDDEN_COMPONENTS = set('habitat2')
ERROR_LIST = []
EMAIL_REGEX = r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b'
GOOGLE_FORM_URL = "https://docs.google.com/forms/d/e/1FAIpQLSfif1hRfUfomonuhJVku7gwqI5L2Wb-D7NzreuU_eiNfchH1g/formResponse?usp=pp_url&entry.1488105878={name}&entry.2089583672={email}"

### Information
def log_parameters(metadata_list, domains, subset, split, components, dest, dest_compressed, ignore_checksum, **kwargs):
  header('-------------------------------------')
  checksum = f'{bcolors.WARNING}False{bcolors.ENDC}' if ignore_checksum else f'{bcolors.OKGREEN}True{bcolors.ENDC}'
  header(f'From {bcolors.OKGREEN}SERVERS{bcolors.ENDC}: (using checksum validation: {checksum})')
  for rsm in metadata_list: header(f'    {bcolors.UNDERLINE}{rsm.link_file}{bcolors.ENDC}')
  header('')
  header(f'Data {bcolors.OKGREEN}parameters{bcolors.ENDC}: (what to download)') 
  header(f'    {bcolors.WARNING}Domains{bcolors.ENDC}    = {domains}') 
  header(f'    {bcolors.WARNING}Components{bcolors.ENDC} = {components}') 
  header(f'    {bcolors.WARNING}Subset{bcolors.ENDC}     = {subset}') 
  header(f'    {bcolors.WARNING}Split{bcolors.ENDC}      = {split}') 
  header('')
  header(f'Data {bcolors.OKGREEN}locations{bcolors.ENDC}:') 
  header(f'    {bcolors.WARNING}Dataset (extracted){bcolors.ENDC}      = {dest}') 
  header(f'    {bcolors.WARNING}Compressed files   {bcolors.ENDC}      = {dest_compressed}') 
  header('-------------------------------------\n\n')
  # print(f'[{bcolors.OKGREEN}FETCHING{bcolors.ENDC}] metadata from:')

def end_notes(**kwargs):
  notice(f'[{bcolors.OKGREEN + bcolors.BOLD}Download complete{bcolors.ENDC}]')
  notice(f'    Number of model files downloaded={len(kwargs["models"])}')
  if len(kwargs['errors']) > 0:
    notice(f'    Ran into {len(kwargs["errors"])} non-critical failures:')
    for err in kwargs['errors']:
      notice(err)
  notice('Recap:')
  log_parameters(**kwargs)

##
def email_valid(email): return re.fullmatch(EMAIL_REGEX, email)
def prompt_email():
  while True: 
    email = input("Please enter a valid email to associate with this download: ").lower()
    if email_valid(email): break
    failure(f'{email} not a valid email.')
  return email

def prompt_name(email):
  while True: 
    res = input(f"Please enter your name associated '{email}': ").lower()
    if res != '': break
  return res

### Pre-download validation
def licenses_clickthrough(components, require_prompt, component_to_license, email, name):
  components = set(components + ['omnidata']) # Make sure everyone accepts omnidata terms
  license('Before continuing the download, please review the terms of use for each of the following component datasets:')
  for component in components:
    license(f"    {bcolors.WARNING}{component}{bcolors.ENDC}: \x1B]8;;{component_to_license[component]}\x1B\\{component_to_license[component]}\x1B]8;;\x1B\\")
  if not require_prompt:
    if not (name and email_valid(email)): raise ValueError("In order to use --agree_all you must also supply a name and valid email through the args --name NAME and --email USER@DOMAIN)")
    print(name, email)
    notice("Confirmation supplied by option '--agree_all'\n")
  else: 
    while True: 
      res = input("By entering 'y', I confirm that I have read and accept the above linked terms and conditions [y/n]: ").lower()
      if res == 'y': break
      elif res == 'n': print(f'[{bcolors.FAIL + bcolors.BOLD}EXIT{bcolors.ENDC}] Agreement declined: cancelling download.'); exit(0)
    if not email: email = prompt_email()
    if not name: name = prompt_name(email)
    notice("Agreement accepted. Continuing download.\n")
  result = urllib.request.urlopen(GOOGLE_FORM_URL.format(name=name, email=email))
  return

def validate_checksums_exist(models):
  models_without_checksum = [m for m in models if m.checksum is None]
  if len(models_without_checksum) > 0: 
    show_k = 100
    notice(f'Found {len(models_without_checksum)} models without checksums:')
    for m in models_without_checksum[:show_k]: print(f'    {m.url}')
    if len(models_without_checksum) > show_k:  print(f'    and {len(models_without_checksum) - show_k} more...')
    print(f'Since "--ignore_checksum=False", cannot continue. Aborting.')
    exit(1)

def filter_models(models, domains, subset, split, components, component_to_split, component_to_subset):
  # for m in models:
  #   if m.component_name.lower() == 'hypersim':
  #     notice(f'{m.component_name} [{m.domain}]')
  #     print(components, domains)
  #     print(m.component_name.lower() in components)
  #     print(subset == 'all' or component_to_subset[m.component_name.lower()] is None or m.model_name in component_to_subset[m.component_name.lower()][subset]) 
  #     print(split == 'all' or component_to_split[m.component_name.lower()] is None or m.model_name in component_to_split[m.component_name.lower()])
  #     print('all' in domains or m.domain in domains)
  #     print("\n")
  # notice(f"Servers contain {len(models)} models")
  # unique_components = set([m.component_name for m in models])
  # component_to_domains = {k: set() for k in unique_components}
  # for m in models: component_to_domains[m.component_name].add(m.domain)
  # notice(f'{unique_components}')
  # for component, found_domains in component_to_domains.items(): notice(f'{component}: {found_domains}')
  print(domains)
  filtered = [m for m in models 
    if (m.component_name.lower() in components)
    and (subset == 'all' or component_to_subset[m.component_name.lower()] is None or m.model_name in component_to_subset[m.component_name.lower()][subset]) 
    and (split == 'all' or component_to_split[m.component_name.lower()] is None or m.model_name in component_to_split[m.component_name.lower()])
    and ('all' in domains or m.domain in domains)
    ]
  notice(f"Filtered down to {len(filtered)} models based on specified criteria.")
  # exit(0)
  return filtered
## 

### Downloading
def ensure_aria2_server(aria2_create_server, aria2_uri, aria2_secret, connections_total, connections_per_server_per_download, aria2_cmdline_opts, **kwargs):
  if not aria2_uri or not aria2_create_server: return None
  a2host, a2port = ":".join(aria2_uri.split(':')[:-1]), aria2_uri.split(':')[-1]
  notice(f"Opening aria2c download daemon in background: {bcolors.WARNING}Run {bcolors.OKCYAN}'aria2p'{bcolors.WARNING} in another window{bcolors.ENDC} to view status.") 
  n = connections_total 
  x = connections_per_server_per_download if connections_per_server_per_download is not None else connections_total
  x = min(x, 16)
  a2server = Popen(('aria2c --enable-rpc --rpc-listen-all --disable-ipv6 -c --auto-file-renaming=false ' +
                    # '--optimize-concurrent-downloads ' + 
                    f'-s{n}  -j{n}  -x{x} {aria2_cmdline_opts}').split())
  atexit.register(os.kill, a2server.pid, signal.SIGINT)
  return aria2p.API(aria2p.Client(host=a2host, port=a2port, secret=aria2_secret))


def get_tar_fname_and_fpath(url, output_dir, output_name=None):
  fname = url.split('/')[-1] if output_name is None else output_name
  fpath = os.path.join(output_dir, fname)
  return fname, fpath

def download_tar(url, output_dir='.', output_name=None, n=20, n_per_server=10,
  checksum=None, max_tries_per_model=3, aria2api=None, dryrun=False,
  ) -> Optional[str]:
  '''Downloads "url" to output filename. Returns downloaded fpath.'''
  fname, fpath = get_tar_fname_and_fpath(url, output_dir, output_name)
  if dryrun: print(f'Downloading "{url}"" to "{fpath}"'); return fpath
  # checksum = checksum[:-3] + '000'
  # print(checksum)
  if aria2api is not None:
    options_dict = { 'out': fname, 'dir': output_dir, 'check_integrity': True}
    if checksum is not None: options_dict['checksum'] = f"md5={checksum}"
    while (max_tries_per_model := max_tries_per_model-1) > 0:
      res = aria2api.client.add_uri(uris=[url], options=options_dict)
      success = wait_on(aria2api, res)
      if success: break
    if not success: return None
  else:
    # os.makedirs(output_dir, exist_ok=True)
    # cmd = f'lftp -e "pget -n {n} {url} -o {fpath}"'
    # # print(cmd)
    # call(cmd, shell=True) 
    options = f'-c --auto-file-renaming=false'
    if n_per_server is None: n_per_server = min(n, 16)
    options += f' -s {n} -j {n} -x {n_per_server}' # N connections
    if checksum is not None: options += f' --check-integrity=true --checksum=md5={checksum}'
    cmd = f'aria2c -k 1M -d {output_dir} -o {fname} {options} "{url}"'
    # print(cmd)
    call(cmd, shell=True) 

    # os.makedirs(output_dir, exist_ok=True)
    # cmd = f'axel -q -o {fpath} -c -n {n} "{url}" '
    # success = True
    # while (max_tries_per_model := max_tries_per_model-1) > 0:
    #   call(cmd, shell=True)
    #   if checksum is not None: success = (check_output(['md5sum', fpath], encoding='UTF-8').split()[0] == checksum)
    # if not success: return None
  return fpath

def wait_on(a2api, gid, duration=0.2):
  while not (a2api.get_downloads([gid])[0].is_complete or a2api.get_downloads([gid])[0].has_failed):
    time.sleep(duration)
  success = a2api.get_downloads([gid])[0].is_complete 
  a2api.remove(a2api.get_downloads([gid]))
  return success
##


### Untarring
def untar(fpath, model, dest=None, ignore_existing=True,
    output_structure=('domain', 'component_name', 'model_name'), # Desired directory structure
    dryrun=False
  ) -> None:
  dest_fpath = os.path.join(dest, *[getattr(model, a) for a in output_structure])
  if dest is not None: os.makedirs(dest, exist_ok=True)
  if os.path.exists(dest_fpath) and ignore_existing: notice(f'"{dest_fpath}" already has some uncompressed files and no .tar file. Assuming this means download was successful and a previous run deleted the compressed file, so skipping re-download. If this behavior is undesired, delete or move this folder'); return
  with tempfile.TemporaryDirectory(dir=dest) as tmpdirname:
    src_fpath = os.path.join(tmpdirname, *[getattr(model, a) for a in model.tar_structure])
    if dryrun: print(f'Extracting "{fpath}"" to "{tmpdirname}" and moving "{src_fpath}" to "{dest_fpath}"'); return
    with tarfile.open(fpath) as tar:
      tar.extractall(path=tmpdirname)
    shutil.move(src_fpath, dest_fpath)

##

@call_parse
def download(
  domains:     Param("Domains to download (comma-separated or 'all')", str, nargs='+'),
  subset:      Param("Subset to download", str, choices=['all', 'debug', 'tiny', 'medium', 'full', 'fullplus'])='debug',
  split:       Param("Split to download", str, choices=['train', 'val', 'test', 'all'])='all',
  components:  Param("Component datasets to download (comma-separated)", str, nargs='+',
    choices=['all','replica','taskonomy','replica_gso','hypersim','blendedmvg','hm3d','clevr_simple','clevr_complex'])='all',
  dest:             Param("Where to put the uncompressed data", str)='uncompressed/',
  dest_compressed:  Param("Where to download the compressed data", str)='compressed/',
  keep_compressed:  Param("Don't delete compressed files after decompression", bool_arg)=False,
  only_download:    Param("Only download compressed data", bool_arg)=False,
  max_tries_per_model:    Param("Number of times to try to download model if checksum fails.", int)=3,  
  connections_total:      Param("Number of simultaneous aria2c connections overall (note: if not using the RPC server, this is per-worker)", int)=32,
  connections_per_server_per_download: Param("Number of simulatneous aria2c connections per server per download. Defaults to 'total_connections' (note: if not using the RPC server, this is per-worker)", int)=None,
  n_workers:              Param("Number of workers to use", int)=min(mp.cpu_count(), 16),
  num_chunk:        Param("Download the kth slice of the overall dataset", int)=0,
  num_total_chunks: Param("Download the dataset in N total chunks. Use with '--num_chunk'", int)=1, 
  ignore_checksum:  Param("Ignore checksum validation", bool_arg)=True,
  dryrun:           Param("Keep compressed files even after decompressing", store_true)=False,
  aria2_uri:              Param("Location of aria2c RPC (if None, use CLI)", str)="http://localhost:6800", 
  aria2_cmdline_opts:     Param("Opts to pass to aria2c", str)='',  
  aria2_create_server:    Param("Create a RPC server at aria2_uri", bool_arg)=True, 
  aria2_secret:           Param("Secret for aria2c RPC", str)='',
  agree_all:      Param("Agree to all license clickwraps.", store_true)=False, 
  email:          Param("Email used to agree to clickwraps.", str)='',
  name:           Param("Name used to agree to clickwraps.", str)='',
  ):
  ''' 
    Downloads Omnidata starter dataset.
    ---
    The data is stored on the remote server in a compressed format (.tar.gz).
    This function downloads the compressed and decompresses it.

    Examples:
      download rgb normals point_info --components clevr_simple clevr_complex --connections_total 30
  '''
  # The following data could instead be supplied from the remote server:
  metadata_list = STARTER_DATASET_REMOTE_SERVER_METADATAS # Param("Metadata servers to search", Iterable[RemoteStorageMetadata])
  component_to_split = STARTER_DATA_COMPONENT_TO_SPLIT    # Param("Train/Val/Test splits for each component", Dict[str, Dict[str, Iterable[str]]])
  component_to_subset = STARTER_DATA_COMPONENT_TO_SUBSET  # Param("Debug/.../fullplus splits for each component", Dict[str, Iterable[str]])
  component_to_license = STARTER_DATA_LICENSES            # Param("Licenses for each component dataset", Dict[str, Iterable[str]])

  if components == 'all': components = list(component_to_license.keys())
  log_parameters(**locals())
  licenses_clickthrough(components, require_prompt=not agree_all, component_to_license=component_to_license, email=email, name=name)
  aria2 = ensure_aria2_server(**locals())

  # Determine which models to use
  models = [metadata.parse(url)
            for metadata in metadata_list 
            for url in metadata.links]
  models = filter_models(models, domains, subset, split, components, 
            component_to_split=component_to_split,
            component_to_subset=component_to_subset)
  notice(f'Found {len(models)} matching blobs on remote serverss.')
  models = models[num_chunk::num_total_chunks] # Parallelization: striped slice of models array
  if not ignore_checksum: validate_checksums_exist(models)


  # Process download

  def process_model(model, ignore_existing=True, output_structure=('domain', 'component_name', 'model_name')):
    try:
      dest_fpath = os.path.join(dest, *[getattr(model, a) for a in output_structure])
      _, tar_fpath = get_tar_fname_and_fpath(model.url, output_dir=dest_compressed, output_name=model.fname)
      if os.path.exists(dest_fpath) and (not os.path.exists(tar_fpath)) and ignore_existing: notice(f'"{dest_fpath}" already has some files... skipping re-download. If this behavior is undesired, delete or move this folder'); return
      tar_fpath = download_tar(
                  model.url, output_dir=dest_compressed, output_name=model.fname, 
                  checksum=None if ignore_checksum else model.checksum,
                  n=connections_total, n_per_server=connections_per_server_per_download,
                  aria2api=aria2, dryrun=dryrun)
      if tar_fpath is None: return
      if only_download:     return
      untar(tar_fpath, dest=dest, model=model, ignore_existing=ignore_existing, dryrun=dryrun, output_structure=output_structure)
      if not keep_compressed: os.remove(tar_fpath)
    except ReadError as e:
      msg = f"ReadError when untarring {model.url}"
      failure(msg)
      return msg
    except Exception as e:
      failure(f"Uncaught error when processing model {model.url} (stacktrace below)")
      raise e

  if n_workers < 1 : 
    errors = [process_model(model) for model in tqdm.tqdm(models)]
  else:
    with mp.Pool(n_workers) as p:
      errors = list(tqdm.tqdm(p.imap(process_model, models), total=len(models)))
      # p.map(process_model, models)
  errors = [f"        {e}" for e in errors if e is not None] 
  
  # Cleanup
  end_notes(**locals())


if __name__ == '__main__':
  a2server = Popen('aria2c --enable-rpc --rpc-listen-all --disable-ipv6 -c --auto-file-renaming=false -s 10 -x 10'.split())

  time.sleep(0.2)
  model = ZippedModel(
    component_name='taskonomy', domain='point_info', model_name='yscloskey', 
    url='https://datasets.epfl.ch/taskonomy/yscloskey_point_info.tar'
  )
  tar_format = ('domain',)
  dest_compressed = '/tmp/omnidata/compressed'
  dest = '/tmp/omnidata/uncompressed'

  tar_fpath = download_tar(model.url, output_dir=dest_compressed,
    checksum='md5=9f9752d74b07bcc164af4a6c61b0eca1',
    output_name=model.fname)
  untar(tar_fpath, dest=dest, model=model, tar_format=tar_format, ignore_existing=True)
  os.remove(tar_fpath)

  # Terminate the process
  os.kill(a2server.pid, signal.SIGINT)




from collections import namedtuple
import functools
from fastcore.basics import store_attr
import requests
from   typing import Any, Optional, Dict, List, Iterable, Tuple, Union, Callable



class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    ENDC = '\033[0m'

def notice(msg): print(f'[{bcolors.OKGREEN + bcolors.BOLD}NOTICE{bcolors.ENDC}] {msg}')
def header(msg): print(f'[{bcolors.HEADER + bcolors.BOLD}HEADER{bcolors.ENDC}] {msg}')
def license(msg): print(f'[{bcolors.WARNING + bcolors.BOLD}LICENSE{bcolors.ENDC}] {msg}')
def underline(msg): print(f'{bcolors.UNDERLINE}{msg}{bcolors.ENDC}')
def failure(msg): print(f'[{bcolors.FAIL + bcolors.BOLD}FAILURE{bcolors.ENDC}] {msg}')

def print_and_log_failure(msg, error_list):
    failure(msg)
    error_list.append(msg)

class ZippedModel:
  def __init__(self, component_name, domain, model_name, url, tar_structure, checksum=None):
    self.component_name = component_name
    self.domain         = domain.lower()
    self.model_name     = model_name
    self.url            = url
    self.ext = ".".join(self.url.split('/')[-1].split('.')[1:])
    self.fname = f'{domain}__{component_name}__{model_name}.{self.ext}'
    self.checksum = checksum
    self.tar_structure = tar_structure

class RemoteStorageMetadata:
  ''' Contains/gets metadata about what/where data is stored on some remote endpoint. '''
  def __init__(self, 
      link_file, checksum_file=None, expected_suffix='.tar', 
      tar_structure=('domain', 'component_name', 'model_name')
    ) -> None: store_attr()
  
  def parse(self, url) -> ZippedModel: raise NotImplementedError
  def checksum(self, url) -> str: return self.checksums[url]
  @functools.cached_property  
  def links(self)      -> List[str]:  return [k for k in requests.get(self.link_file).text.splitlines() if k.endswith(self.expected_suffix)]
  @functools.cached_property  
  def checksums(self)    -> Dict[str, str]:  return {k.split()[1]: k.split()[0] for k in requests.get(self.checksum_file).text.splitlines() if k.endswith(self.expected_suffix)}
  @functools.cached_property  
  def models(self)     -> List[ZippedModel]:  return [self.parse(url) for url in self.links]
  @functools.cached_property  
  def domains(self)    -> List[str]:  return set([m.domain for m in self.models])
  @functools.cached_property  
  def components(self) -> List[str]:  return set([m.component_name for m in self.models])
  @functools.cached_property 
  def info(self)       -> str:  return join_recursive('\n', [
      f'Data location: {self.link_file}',
      f'    Links: ({len(self.links)})',
      f'    Domains: ({len(self.domains)})',
      [f'      {d}' for d in self.domains],
      f'    Components: ({len(self.components)})',
      [f'      {c}' for c in self.components]
    ])

class RemoteBucketStorageMetadata(RemoteStorageMetadata):
  '''
    Remote storage, but assumes there are two files at:
      BASE_URL/links.txt
      BASE_URL/md5sum.txt
    
    - links.txt should describe the full url download link for each model
    - md5sum.txt should be two-column with a row for each model: first col is md5sum 
        and second col is the part of the URL specific to the model (e.g. everything after *.edu/)
  '''
  def __init__(self, base_url, expected_suffix='.tar.gz', tar_structure=('domain', 'component_name', 'model_name')) -> None:
    self.base_url = base_url
    super().__init__(
      link_file=f'{base_url}/links.txt', checksum_file=f'{base_url}/md5sum.txt',
      expected_suffix=expected_suffix, tar_structure=tar_structure
    )

  def checksum(self, url): return self.checksums.get(url.replace(self.base_url,''))

def join_recursive(s, args):
  return s.join([x if isinstance(x, str) else join_recursive(s, x) for x in args])



from   dataclasses import dataclass, field
from   typing import Optional, List, Callable, Union, Dict, Any
import os, sys
import os
import multiprocessing as mp
from ..splits import get_splits, get_all_spaces

###############################################################################
# Split info:
# Exports:
#   e.g. flat_split_to_spaces['tiny-train'] -> List[str] building names
###############################################################################
subsets = ['debug', 'tiny', 'medium', 'full', 'fullplus']

forbidden_buildings = ['mosquito', 'tansboro', 'tomkins', 'darnestown', 'brinnon', 'rough', 'woodbine'] # We do not have the rgb data for tomkins, darnestown, brinnon
# forbidden_buildings += ['rough'] # Contains some wrong viewpoints
# forbidden_buildings += ['wiconisco'] # Bad texture?
# forbidden_buildings = []

taskonomy_split_files = {
        s: os.path.join( os.path.dirname(__file__), 'train_val_test_{}.csv'.format(s.lower()) )
        for s in subsets
}

# e.g. taskonomy_split_to_buildings['tiny']['train'] -> List[str] building names
taskonomy_split_to_buildings = {
    s: get_splits(taskonomy_split_files[s], forbidden_buildings=forbidden_buildings) 
    for s in subsets
}

# e.g. flat_split_to_spaces['tiny-train'] -> List[str] building names
flat_split_to_spaces = {}
for subset in taskonomy_split_to_buildings:
    for split, buildings in taskonomy_split_to_buildings[subset].items():
        flat_split_to_spaces[subset + '-' + split] = buildings

split_to_spaces = taskonomy_split_to_buildings['fullplus']
subset_to_spaces = {subset: get_all_spaces(splits) for subset, splits in taskonomy_split_to_buildings.items()}
###############################################################################
#  Semantic segmentation:
###############################################################################

CLASS_LABELS = [
    '__background__', 'bicycle', 'car', 'motorcycle',
    'boat', 'bench', 'backpack',
    'umbrella', 'handbag', 'tie', 'suitcase', 'frisbee', 'skis',
    'snowboard', 'sports ball', 'kite', 'baseball bat', 'baseball glove',
    'skateboard', 'surfboard', 'tennis racket', 'bottle', 'wine glass',
    'cup', 'fork', 'knife', 'spoon', 'bowl', 'banana', 'apple', 'sandwich',
    'orange', 'broccoli', 'carrot', 'hot dog', 'pizza', 'donut', 'cake',
    'chair', 'couch', 'potted plant', 'bed', 'dining table', 'toilet', 'tv',
    'laptop', 'mouse', 'remote', 'keyboard', 'cell phone', 'microwave',
    'oven', 'toaster', 'sink', 'refrigerator', 'book', 'clock', 'vase',
    'scissors', 'teddy bear', 'hair drier', 'toothbrush'
]



import csv
import os

forbidden_buildings = ['mosquito', 'tansboro']

# The 'small-image' dataset has some misssing buildings. Either we can ignore those buildings or we can regenerate the dataset.
forbidden_buildings = ['newfields', 'ludlowville', 'goodyear', 'castroville']  # missing RGB
# forbidden_buildings = ['mosquito', 'tansboro', 'tomkins', 'darnestown', 'brinnon']
# We do not have the rgb data for tomkins, darnestown, brinnon

forbidden_buildings += ['rough'] # Contains some wrong view-points

# SPLIT_TO_NUM_IMAGES = {
#     'supersmall': 14575,
#     'tiny': 262745,
#     'fullplus': 3349691,
# }

def get_splits(split_path):
    with open(split_path) as csvfile:
        readCSV = csv.reader(csvfile, delimiter=',')

        train_list = []
        val_list = []
        test_list = []

        for row in readCSV:
            name, is_train, is_val, is_test = row
            if name in forbidden_buildings:
                continue
            if is_train == '1':
                train_list.append(name)
            if is_val == '1':
                val_list.append(name)
            if is_test == '1':
                test_list.append(name)
    return {
        'train': sorted(train_list),
        'val': sorted(val_list),
        'test': sorted(test_list)
    }


# Taskonomy
subsets = ['debug', 'tiny', 'medium', 'full', 'fullplus']
taskonomy_split_files = {s:  os.path.join(os.path.dirname(__file__),
                                'splits',
                                'train_val_test_{}.csv'.format(s.lower()))
               for s in subsets}

taskonomy_split_to_buildings = {s: get_splits(taskonomy_split_files[s]) for s in subsets}

taskonomy_flat_split_to_buildings = {}
for subset in taskonomy_split_to_buildings:
    for split, buildings in taskonomy_split_to_buildings[subset].items():
        taskonomy_flat_split_to_buildings[subset + '-' + split] = buildings


# Replica
replica_split_file = os.path.join(os.path.dirname(__file__), 'splits', 'train_val_test_replica.csv')
replica_flat_split_to_buildings = get_splits(replica_split_file)

# GSO
gso_split_file = os.path.join(os.path.dirname(__file__), 'splits', 'train_val_test_gso.csv')
gso_flat_split_to_buildings = get_splits(gso_split_file)

# hypersim
hypersim_split_file = os.path.join(os.path.dirname(__file__), 'splits', 'train_val_test_hypersim.csv')
hypersim_flat_split_to_buildings = get_splits(hypersim_split_file)

# blendedMVS
blendedMVS_split_file = os.path.join(os.path.dirname(__file__), 'splits', 'train_val_test_blendedMVS.csv')
blendedMVS_flat_split_to_buildings = get_splits(blendedMVS_split_file)




import os, sys, json
from ..splits import get_splits, get_all_spaces

###############################################################################
# Split info:
# Exports:
#   e.g. flat_split_to_spaces['tiny-train'] -> List[str] building names
###############################################################################
split_file = os.path.join(os.path.dirname(__file__), 'train_val_test_gso.csv')
split_to_spaces = get_splits(split_file)
subset_to_spaces = {'debug': ['frl_apartment_0-3'], 'fullplus': get_all_spaces(split_to_spaces) }

###############################################################################
#  Semantic segmentation:
###############################################################################

REPLICA_CLASS_LABELS = [
    'undefined', 'backpack', 'base-cabinet', 'basket', 'bathtub', 'beam', 'beanbag', 'bed', 'bench', 'bike',
    'bin', 'blanket', 'blinds', 'book', 'bottle', 'box', 'bowl', 'camera', 'cabinet', 'candle', 'chair',
    'chopping-board', 'clock', 'cloth', 'clothing', 'coaster', 'comforter', 'computer-keyboard', 'cup',
    'cushion', 'curtain', 'ceiling', 'cooktop', 'countertop', 'desk', 'desk-organizer', 'desktop-computer',
    'door', 'exercise-ball', 'faucet', 'floor', 'handbag', 'hair-dryer', 'handrail', 'indoor-plant',
    'knife-block', 'kitchen-utensil', 'lamp', 'laptop', 'major-appliance', 'mat', 'microwave', 'monitor',
    'mouse', 'nightstand', 'pan', 'panel', 'paper-towel', 'phone', 'picture', 'pillar', 'pillow', 'pipe',
    'plant-stand', 'plate', 'pot', 'rack', 'refrigerator', 'remote-control', 'scarf', 'sculpture', 'shelf',
    'shoe', 'shower-stall', 'sink', 'small-appliance', 'sofa', 'stair', 'stool', 'switch', 'table',
    'table-runner', 'tablet', 'tissue-paper', 'toilet', 'toothbrush', 'towel', 'tv-screen', 'tv-stand',
    'umbrella', 'utensil-holder', 'vase', 'vent', 'wall', 'wall-cabinet', 'wall-plug', 'wardrobe', 'window',
    'rug', 'logo', 'bag', 'set-of-clothing'
]

GSO_CLASS_LABELS = [
    'undefined', 'backpack', 'bag', 'bottle', 'cup', 'spoon', 'bowl', 'laptop','mouse', 
    'keyboard', 'toaster', 'clock', 'scissors', 'hair drier', 'basket', 'bin', 'box', 
    'camera', 'desk-organizer', 'kitchen-utensil', 'pan', 'plate', 'shoe', 'small-appliance', 
    'tablet', 'towel', 'utensil-holder', 'boot', 'hat', 'pencil case', 'teapot', 'headphones', 
    'hammer', 'gloves', 'tape', 'file sorter', 'mug', 'slipper', 'pill bottle', 'coffee maker', 
    'hard drive', 'speaker', 'cable', 'pitcher', 'flower pot', 'stuffed animal', 'plastic animal', 
    'toy', 'bucket', 'sponge', 'game console', 'pet food bowl', 'flashlight', 'video game', 
    'USB flash drive', 'dustpan', 'hair straightener', 'dish drainer', 'tube', 'can', 
    'food package', 'medicine package', 'game package', 'random package'
]

REPLICA_LABEL_TRANSFORM = [
    0, 6, 62, 63, 64, 0, 65, 41, 5, 1, 66, 67, 68, 55, 21, 69, 27, 70, 62, 71, 38, 72, 56, 73, 74, 
    75, 76, 48, 23, 77, 78, 79, 80, 81, 82, 83, 84, 85, 14, 86, 87, 8, 60, 88, 40, 25, 89, 90, 45,
    91, 92, 50, 93, 46, 94, 95, 0, 96, 97, 98, 0, 99, 0, 40, 100, 101, 102, 54, 47, 103, 104, 102,
    105, 106, 53, 107, 39, 108, 109, 110, 42, 73, 111, 96, 43, 61, 112, 44, 113, 7, 114, 57, 115, 116, 62, 
    117, 118, 119, 120, 121, 8, 74
]

gso_label_transform_file = os.path.join(os.path.dirname(__file__), 'GSO_LABEL_TRANSFORM.json')
with open(gso_label_transform_file) as file:
    GSO_LABEL_TRANSFORM = json.load(file)


###################
# GSO + replica 
###################
# class = 2**8 * r + g
# instance = b 
# number of classes : 102 (replica) + 1032 (google objects)
# class < 102 -> class in replica
# class >= 102 -> (class - 102) in GSO


GSO_NUM_CLASSES = len(REPLICA_CLASS_LABELS) + 1032

###############################################################################
# Make dataset 
###############################################################################

REPLICA_BUILDINGS = [
    'apartment_0',
    'apartment_1', 
    'apartment_2',
    'frl_apartment_0',
    'frl_apartment_1',
    'frl_apartment_2',
    'frl_apartment_3',
    'frl_apartment_4',
    'frl_apartment_5',
    'office_0',
    'office_1',
    'office_2',
    'office_3',
    'office_4',
    'room_1'
    'room_0',
    'room_2',
    'hotel_0',
]